import Bio.Phylo
import copy


def prune_leaves(tree, f):
    """ Prune leaves from a tree based on a predicate

    Parameters
    ----------
    tree : Bio.Phylo.BaseTree.Tree
        Tree to prune
    f : function
        Predicate function that takes a clade and returns a boolean
    """

    # Generate new clade object
    clade = prune_leaves_recursive(tree.clade, f)

    # Create a new Tree with the pruned clade as its root
    pruned_tree = Bio.Phylo.BaseTree.Tree(root=clade)

    return pruned_tree


def prune_leaves_recursive(clade, f):
    """ Prune leaves from a tree based on a predicate
    """
    # Create a copy of the clade
    clade = copy.deepcopy(clade)

    # If the clade is a leaf, check the predicate
    if clade.is_terminal():
        if f(clade):
            return None
        else:
            return clade

    # If the clade is not a leaf, recurse on its children
    else:
        new_clades = []
        for a in clade.clades:
            new_a = prune_leaves_recursive(a, f)
            if new_a is not None:
                new_clades.append(new_a)

        if len(new_clades) == 0:
            return None
        else:
            clade.clades = new_clades
            return clade


def aggregate_tree_branches(tree, f_merge=None):
    """ Merge redundant branches of a tree

    Parameters
    ----------
    tree : Bio.Phylo.BaseTree.Tree
        Tree to merge
    f_merge : function
        Function to specify how to merge branches. If None, the default of summming branch lengths is used

    Returns
    -------
    Bio.Phylo.BaseTree.Tree
        Tree with merged branches
    """
    if f_merge is None:
        f_merge = lambda parent, child: {'branch_length': parent.branch_length + child.branch_length}

    # Generate new clade object
    clade = aggregate_tree_branches_recursive(tree.clade, f_merge)

    # Create a new Tree with the pruned clade as its root
    agg_tree = Bio.Phylo.BaseTree.Tree(root=clade)

    return agg_tree


def aggregate_tree_branches_recursive(clade, f_merge):
    """ Merge redundant branches of a tree

    Merge chains of branches into a single branch, aggregating properties of the branches.

    Parameters
    ----------
    clade : Bio.Phylo.BaseTree.Clade
        Clade to merge
    f_merge : function
        Function to specify how to merge branches

    Returns
    -------
    Bio.Phylo.BaseTree.Clade
        Clade with merged branches
    """

    # If the clade is not a leaf and has 1 child, merge with child
    if len(clade.clades) == 1:
        # Recurse on child
        new_child_clade = aggregate_tree_branches_recursive(clade.clades[0], f_merge)

        # Merge clade and child properties
        properties = f_merge(clade, new_child_clade)

        # Create new clade with given properties
        new_merged_clade = Bio.Phylo.BaseTree.Clade(
            clades=new_child_clade.clades,
            branch_length=properties.get('branch_length'),
            name=properties.get('name'),
            confidence=properties.get('confidence'),
            color=properties.get('color'),
            width=properties.get('width'))

        for key, value in properties.items():
            if key not in ['branch_length', 'name', 'confidence', 'color', 'width']:
                setattr(new_merged_clade, key, value)

        return new_merged_clade

    else:
        # If the clade is not a leaf and has more than 1 child, recurse on each child
        new_clades = []
        for a in clade.clades:
            new_a = aggregate_tree_branches_recursive(a, f_merge)
            new_clades.append(new_a)

        # Create a new clade with the new children
        clade = copy.deepcopy(clade)
        clade.clades = new_clades

        return clade


def align_cn_tree(tree, adata):
    """ Align a tree's leaves with an AnnData object's observations

    Parameters
    ----------
    tree : Bio.Phylo.BaseTree.Tree
        Tree to align
    adata : anndata.AnnData
        AnnData object to align tree with

    Returns
    -------
    Bio.Phylo.BaseTree.Tree
        Aligned tree
    anndata.AnnData
        Aligned AnnData object
    """
    tree = prune_leaves(tree, lambda a: a.name not in adata.obs.index)

    cell_ids = []
    for a in tree.get_terminals():
        cell_ids.append(a.name)

    adata = adata[cell_ids]

    return tree, adata



import logging
import sklearn.cluster
import sklearn.mixture
import sklearn.preprocessing
import umap
import pandas as pd
import numpy as np
import anndata as ad
from natsort import natsorted

from anndata import AnnData
from typing import Dict, Any, Union, Iterable

import scgenome.cncluster
import scgenome.preprocessing.transform


def _kmeans_bic(X, k):
    model = sklearn.cluster.KMeans(n_clusters=k, init='k-means++', random_state=100).fit(X)
    bic = scgenome.cncluster.compute_bic(model, X)
    labels = model.labels_

    return labels, bic


def _gmm_diag_bic(X, k):
    model = sklearn.mixture.GaussianMixture(n_components=k, covariance_type='diag', init_params='kmeans', random_state=100)
    labels = model.fit_predict(X)
    bic = model.bic(X)

    return labels, bic


def cluster_cells(
        adata: AnnData,
        layer_name: Union[None, str, Iterable[Union[None,str]]]='copy',
        method: str='kmeans_bic',
        min_k: int=2,
        max_k: int=100,
        cell_ids: Iterable[str]=None,
        bin_ids: Iterable[str]=None,
        standardize: bool=False,
    ) -> AnnData:
    """ Cluster cells by copy number.

    Parameters
    ----------
    adata : AnnData
        copy number data
    layer_name : str, optional
        layer with copy number data to plot, None for X, by default 'state'
    method : str, optional
        clustering method, by default 'kmeans_bic'
    min_k : int, optional
        minimum number of clusters, by default 2
    max_k : int, optional
        maximum number of clusters, by default 100
    cell_ids : str, optional
        subset of cells to cluster, by default None
    bin_ids : str, optional
        subset of bins to cluster, by default None
    standarize : bool
        standardize the data prior to outlier detection, by default False
    Returns
    -------
    AnnData
        copy number data with additional `cluster_id` and `cluster_size` columns

    Examples
    -------

    >>> import scgenome
    >>> import anndata as ad
    >>> import numpy as np
    >>> adata = ad.AnnData(np.array([
    ...    [3, 3, 3, 6, 6],
    ...    [1, 1, 1, 2, 2],
    ...    [1, 22, 1, 2, 2],
    ...    [1, 3, 3, 5, 5],
    ... ]).astype(np.float32))
    >>> adata = scgenome.tl.cluster_cells_kmeans(adata, layer_name=None, max_k=3)
    >>> adata.obs['cluster_id']
    0    0
    1    2
    2    1
    3    0
    Name: cluster_id, dtype: category
    Categories (3, int64): [0, 1, 2]

    """
    if cell_ids is None:
        cell_ids = adata.obs.index

    if bin_ids is None:
        bin_ids = adata.var.index

    min_k = min(adata.shape[0], min_k)
    max_k = min(adata.shape[0], max_k)

    def __get_layer(layer_name):
        if layer_name is not None:
            return np.array(adata[cell_ids, bin_ids].layers[layer_name])
        else:
            return np.array(adata[cell_ids, bin_ids].X)

    if isinstance(layer_name, (str, type(None))):
        X = __get_layer(layer_name)
    elif isinstance(layer_name, Iterable):
        X = np.concatenate([__get_layer(l) for l in layer_name], axis=1)
    else:
        raise ValueError(f'layer_name was {layer_name}')

    X = scgenome.preprocessing.transform.fill_missing(X)
    
    if standardize:
        X = sklearn.preprocessing.StandardScaler().fit_transform(X)

    ks = range(min_k, max_k + 1)

    logging.info(f'trying with max k={max_k}')

    labels = []
    criterias = []
    for k in ks:
        logging.info(f'trying with k={k}')
        if method == 'kmeans_bic':
            label, criteria = _kmeans_bic(X, k)
        elif method == 'gmm_diag_bic':
            label, criteria = _gmm_diag_bic(X, k)
        else:
            raise ValueError(f'unrecognized method {method}')
        labels.append(label)
        criterias.append(criteria)

    opt_k_idx = np.array(criterias).argmax()
    opt_k = ks[opt_k_idx]
    opt_label = labels[opt_k_idx]
    logging.info(f'selected k={opt_k}')
    
    adata.obs['cluster_id'] = '-1'
    adata.obs.loc[cell_ids, 'cluster_id'] = pd.Series(opt_label, index=adata.obs.loc[cell_ids].index).astype('str').astype('category')
    adata.obs['cluster_size'] = adata.obs.groupby('cluster_id')['cluster_id'].transform('size')

    # store information on the clustering parameters
    adata.uns['clustering'] = {}
    adata.uns['clustering']['params'] = dict(
        method=method,
        opt_k=opt_k,
        min_k=min_k,
        max_k=max_k,
        layer_name=layer_name,
        cell_ids=np.array(cell_ids),
        bin_ids=np.array(bin_ids),
        standardize=standardize,
    )

    return adata


def detect_outliers(
        adata: AnnData,
        layer_name: Union[None, str, Iterable[Union[None,str]]]='copy',
        method: str='isolation_forest',
        standarize: bool=False,
    ) -> AnnData:
    """ Detect outlier cells by copy number.

    Parameters
    ----------
    adata : AnnData
        copy number data
    layer_name : str, optional
        layer with copy number data to use for outlier detection, None for X, by default 'copy'
    method : str, optional
        outlier method, by default 'isolation_forest'
    standarize : bool
        standardize the data prior to outlier detection, by default False

    Returns
    -------
    AnnData
        copy number data with additional `is_outlier` column

    """
    def __get_layer(layer_name):
        if layer_name is not None:
            return np.array(adata.layers[layer_name])
        else:
            return np.array(adata.X)

    if isinstance(layer_name, str):
        X = __get_layer(layer_name)
    elif isinstance(layer_name, Iterable):
        X = np.concatenate([__get_layer(l) for l in layer_name], axis=1)

    X = scgenome.preprocessing.transform.fill_missing(X)

    if standarize:
        X = sklearn.preprocessing.StandardScaler().fit_transform(X)

    if method == 'isolation_forest':
        model = sklearn.ensemble.IsolationForest(random_state=100)
        is_outlier = (model.fit_predict(X) == -1) * 1
    elif method == 'local_outlier_factor':
        model = sklearn.neighbors.LocalOutlierFactor()
        is_outlier = (model.fit_predict(X) == -1) * 1
    else:
        raise ValueError(f'unknown method {method}')

    adata.obs['is_outlier'] = pd.Series(is_outlier, index=adata.obs.index, dtype='category')

    # store information on the clustering parameters
    adata.uns['outliers'] = {}
    adata.uns['outliers']['params'] = dict(
        method='isolation_forest',
        layer_name=layer_name,
        standarize=standarize,
    )

    return adata


def aggregate_clusters(
        adata: AnnData,
        agg_X: Any=None,
        agg_layers: Dict=None,
        agg_obs: Dict=None,
        cluster_col: str='cluster_id',
        cluster_size_col: str='cluster_size') -> AnnData:
    """ Aggregate copy number by cluster to create cluster CN matrix

    Parameters
    ----------
    adata : AnnData
        copy number data
    agg_X : Any
        function to aggregate X, by default None
    agg_layers : Dict, optional
        functions to aggregate layers keyed by layer names, by default None
    agg_obs : Dict, optional
        functions to aggregate obs data keyed by obs columns, by default None
    cluster_col : str, optional
        column with cluster ids, by default 'cluster_id'
    cluster_size_col : str, optional
        column that will be set to the size of each cluster, by default 'cluster_size'

    Returns
    -------
    AnnData
        aggregated cluster copy number
    """

    if agg_X is not None:
        X = (
            adata
                .to_df()
                .set_index(adata.obs[cluster_col].astype(str))
                .groupby(level=0)
                .agg(agg_X)
                .sort_index())
        dtypes = X.dtypes.unique()
        assert len(dtypes) == 1
        dtype = dtypes[0]

    else:
        X = None
        dtype = None

    layer_data = None
    if agg_layers is not None:
        layer_data = {}
        for layer_name in agg_layers:
            layer_data[layer_name] = (
                adata
                    .to_df(layer=layer_name)
                    .set_index(adata.obs[cluster_col].astype(str))
                    .groupby(level=0)
                    .agg(agg_layers[layer_name])
                    .sort_index())

    obs_data = {}
    obs_data[cluster_size_col] = (
        adata.obs
            .set_index(adata.obs[cluster_col].astype(str))
            .groupby(level=0)
            .size())

    if agg_obs is not None:
        for obs_name in agg_obs:
            obs_data[obs_name] = (
                adata.obs
                    .set_index(adata.obs[cluster_col].astype(str))[obs_name]
                    .groupby(level=0)
                    .agg(agg_obs[obs_name])
                    .sort_index())

    obs_data = pd.DataFrame(obs_data)

    adata = ad.AnnData(
        X,
        dtype=dtype,
        obs=obs_data,
        var=adata.var,
        layers=layer_data,
    )

    return adata


def aggregate_clusters_hmmcopy(adata: AnnData) -> AnnData:
    """ Aggregate hmmcopy copy number by cluster to create cluster CN matrix

    Parameters
    ----------
    adata : AnnData
        hmmcopy copy number data

    Returns
    -------
    AnnData
        aggregsated cluster copy number
    """

    agg_X = np.sum

    agg_layers = {
        'copy': np.nanmean,
        'state': np.nanmedian,
    }

    agg_obs = {
        'total_reads': np.nansum,
    }

    return aggregate_clusters(adata, agg_X, agg_layers, agg_obs, cluster_col='cluster_id')


def compute_umap(
        adata: AnnData,
        layer_name: str='copy',
        n_components: int=2,
        n_neighbors: int=15,
        min_dist: float=0.1,
        metric: str='euclidean',
    ) -> AnnData:
    """ Cluster cells by copy number.

    Parameters
    ----------
    adata : AnnData
        copy number data
    layer_name : str, optional
        layer with copy number data on which to perform umap dimensionality
        reduction, None for X, by default 'copy'
    n_components : int
        umap n_components param
    n_neighbors : int
        umap n_neighbors param
    min_dist : float
        umap min_dist param

    Returns
    -------
    AnnData
        copy number data with additional `umap_1`, `umap_2` columns
    """

    if layer_name is not None:
        X = adata.layers[layer_name]
    else:
        X = adata.X

    X = scgenome.preprocessing.transform.fill_missing(X)

    embedding = umap.UMAP(
        n_neighbors=n_neighbors,
        min_dist=min_dist,
        n_components=n_components,
        metric=metric,
        random_state=42,
    ).fit_transform(X)

    adata.obs['UMAP1'] = embedding[:, 0]
    adata.obs['UMAP2'] = embedding[:, 1]

    return adata



import pyfaidx
import pyBigWig
import numba
import numpy as np
import pandas as pd

import scgenome.refgenome
import scgenome.tools.ranges


def _chromosome_count_gc(data, **kwargs):
    genome_fasta = kwargs['genome_fasta']
    chromosome = kwargs['chromosome']
    column_name = kwargs['column_name']

    genome = pyfaidx.Fasta(genome_fasta, rebuild=False)
    chrom_sequence = np.fromiter(str(genome[chromosome]).upper(), dtype='<U1')

    gc_indicator = ((chrom_sequence == 'G') | (chrom_sequence == 'C')) * 1
    gc_cumsum = gc_indicator.cumsum()

    data[column_name] = gc_cumsum[data['End'].values - 1] - gc_cumsum[data['Start'].values]

    return data


def count_gc(bins, genome_fasta, column_name='gc', proportion=False):
    """ Count gc in each bin

    Parameters
    ----------
    bins : pyranges.PyRanges
        ranges for which to count gc
    genome_fasta : str
        reference genome fasta
    column_name : str, optional
        column to add to `bins`, by default 'gc'
    proportion : bool, optional
        proportion of length, by default False

    Returns
    -------
    pyranges.PyRanges
        output ranges with additional column for gc count
    """    

    data = bins.apply(_chromosome_count_gc, genome_fasta=genome_fasta, column_name=column_name)

    if proportion:
        data = data.assign(column_name, lambda df: df[column_name] / (df['End'] - df['Start']))

    return data


def add_cyto_giemsa_stain(bins):
    """ Add bin specific giesma stain values

    Parameters
    ----------
    bins : pandas.DataFrame
        dataframe with columns 'chr', 'start', 'end' for which to add giesma stain values

    Returns
    -------
    pandas.DataFrame
        output dataframe with columns 'chr', 'start', 'end' and additional column for giesma stain values
    """    

    bins_pr = scgenome.tools.ranges.dataframe_to_pyranges(bins.rename_axis('_index').reset_index())
    cyto_pr = scgenome.tools.ranges.dataframe_to_pyranges(scgenome.refgenome.info.cytobands)

    intersect_1 = bins_pr.intersect(cyto_pr)
    intersect_2 = cyto_pr.intersect(bins_pr)

    intersect = pd.merge(
        scgenome.tools.ranges.pyranges_to_dataframe(intersect_1),
        scgenome.tools.ranges.pyranges_to_dataframe(intersect_2))

    intersect['_width'] = intersect['end'] - intersect['start'] + 1

    selected = intersect.sort_values('_width').drop_duplicates(['_index'], keep='last').set_index('_index')

    cols = ['cyto_band_name', 'cyto_band_giemsa_stain']
    bins = bins.merge(selected[cols], left_index=True, right_index=True, how='left')

    return bins


@numba.jit(nopython=True)
def mean_ranges(values, sequence, starts, ends):
    for i in range(starts.shape[0]):
        values[i] = np.nanmean(sequence[starts[i]:ends[i]])


def _chromosome_mean_bigwig(data, **kwargs):
    bigwig_file = kwargs['bigwig_file']
    chromosome = kwargs['chromosome']
    column_name = kwargs['column_name']
    chr_prefix = kwargs['chr_prefix']

    chromosome = chr_prefix + chromosome
    
    bw = pyBigWig.open(bigwig_file, 'r')

    if chromosome not in bw.chroms():
        data[column_name] = None
        return data

    chromosome_length = bw.chroms()[chromosome]
    sequence = np.array(bw.values(chromosome, start=0, end=chromosome_length))

    x = np.zeros(data.shape[0])
    mean_ranges(x, sequence, data['Start'].values, data['End'].values)

    data[column_name] = x

    return data


def mean_from_bigwig(bins, bigwig_file, column_name, chr_prefix=''):
    """ Count gc in each bin

    Parameters
    ----------
    bins : pyranges.PyRanges
        ranges for which to count gc
    bigwig_file : str
        bigwig filename
    column_name : str
        column to add to `bins`
    chr_prefix : str
        prefix for chromosome names, default ''

    Returns
    -------
    pyranges.PyRanges
        output ranges with additional column for mean bigwig value per bin
    """    

    data = bins.apply(
        _chromosome_mean_bigwig,
        bigwig_file=bigwig_file,
        column_name=column_name,
        chr_prefix=chr_prefix,
    )

    return data




import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from anndata import AnnData
from pandas import DataFrame
from scgenome import refgenome
from scipy.sparse import issparse
from scgenome.plotting import cn_colors
from scgenome.tools.getters import get_obs_data


def genome_axis_plot(data, plot_function, position_columns, **kwargs):
    data = data.merge(refgenome.info.chromosome_info)
    for columns in position_columns:
        data[columns] = data[columns] + data['chromosome_start']

    plot_function(data=data, **kwargs)


def setup_genome_xaxis_lims(ax, chromosome=None, start=None, end=None):
    if chromosome is not None:
        chromosome_start = refgenome.info.chromosome_info.set_index('chr').loc[chromosome, 'chromosome_start']
        chromosome_end = refgenome.info.chromosome_info.set_index('chr').loc[chromosome, 'chromosome_end']

        if start is not None:
            plot_start = chromosome_start + start
        else:
            plot_start = chromosome_start

        if end is not None:
            plot_end = chromosome_start + end
        else:
            plot_end = chromosome_end

    else:
        plot_start = 0
        plot_end = refgenome.info.chromosome_info['chromosome_end'].max()

    ax.set_xlim((plot_start-0.5, plot_end+0.5))


def setup_genome_xaxis_ticks(ax, chromosome=None, start=None, end=None, major_spacing=2e7, minor_spacing=1e6, chromosome_names=None):
    if chromosome_names is None:
        chromosome_names = {}
    
    if chromosome is not None:
        if major_spacing is None:
            major_spacing = 2e7

        if minor_spacing is None:
            minor_spacing = 1e6

        chromosome_length = refgenome.info.chromosome_info.set_index('chr').loc[
            chromosome, 'chromosome_length']
        chromosome_start = refgenome.info.chromosome_info.set_index('chr').loc[chromosome, 'chromosome_start']
        chromosome_end = refgenome.info.chromosome_info.set_index('chr').loc[chromosome, 'chromosome_end']

        xticks = np.arange(0, chromosome_length, major_spacing)
        xticklabels = ['{0:d}M'.format(int(x / 1e6)) for x in xticks]
        xminorticks = np.arange(0, chromosome_length, minor_spacing)

        ax.set_xticks(xticks + chromosome_start)
        ax.set_xticklabels(xticklabels)

        ax.xaxis.set_minor_locator(matplotlib.ticker.FixedLocator(xminorticks + chromosome_start))
        ax.xaxis.set_minor_formatter(matplotlib.ticker.NullFormatter())

    else:
        ax.set_xticks([0] + refgenome.info.chromosome_info['chromosome_end'].values.tolist())
        ax.set_xticklabels([])

        ax.xaxis.set_minor_locator(
            matplotlib.ticker.FixedLocator(refgenome.info.chromosome_info['chromosome_mid'])
        )
        ax.xaxis.set_minor_formatter(matplotlib.ticker.FixedFormatter([chromosome_names.get(c, c) for c in refgenome.info.chromosomes]))


def setup_squash_yaxis(ax):
    squash_coeff = 0.15
    squash_fwd = lambda a: np.tanh(squash_coeff * a)
    squash_rev = lambda a: np.arctanh(a) / squash_coeff
    ax.set_yscale('function', functions=(squash_fwd, squash_rev))

    yticks = np.array([0, 2, 4, 7, 20])
    ax.set_yticks(yticks)

    # Matplotlib will set problematic ylim if there are large y values
    ylim = ax.get_ylim()
    ax.set_ylim(-0.25, ylim[1])
 

def plot_profile(
        data: DataFrame,
        y,
        hue=None,
        ax=None,
        palette=None,
        chromosome=None,
        start=None,
        end=None,
        squashy=False,
        tick_major_spacing=None,
        tick_minor_spacing=None,
        **kwargs
):
    """Plot scatter points of copy number across the genome or a chromosome.

    Parameters
    ----------
    data : pandas.DataFrame
        copy number data
        observation to plot
    y : str
        field with values for y axis
    hue : str, optional
        field by which to color points, None for no color, by default None
    ax : matplotlib.axes.Axes, optional
        existing axess to plot into, by default None
    palette : str, optional
        color palette passed to sns.scatterplot
    chromosome : str, optional
        single chromosome plot, by default None
    start : int, optional
        start of plotting region
    end : int, optional
        end of plotting region
    squashy : bool, optional
        compress y axis, by default False
    rawy : bool, optional
        raw data on y axis, by default False
    tick_major_spacing : int, optional
        major tick spacing, by default 
    tick_minor_spacing : int, optional
        minor tick spacing, by default 1e6
    **kwargs :
        kwargs for sns.scatterplot

    Returns
    -------
    matplotlib.axes.Axes
        Axes used for plotting

    Examples
    -------

    .. plot::
        :context: close-figs

        import scgenome
        adata = scgenome.datasets.OV2295_HMMCopy_reduced()
        scgenome.pl.plot_profile(adata[:, adata.var['gc'] > 0], 'gc')

    """

    if ax is None:
        ax = plt.gca()

    if 'linewidth' not in kwargs:
        kwargs['linewidth'] = 0

    if 's' not in kwargs:
        kwargs['s'] = 5

    if palette is None and hue is not None:
        palette = cn_colors.color_reference

    genome_axis_plot(
        data,
        sns.scatterplot,
        ('start',),
        x='start',
        y=y,
        hue=hue,
        palette=palette,
        ax=ax,
        **kwargs)

    setup_genome_xaxis_ticks(
        ax,
        chromosome=chromosome,
        start=start,
        end=end,
        major_spacing=tick_major_spacing,
        minor_spacing=tick_minor_spacing,
    )

    setup_genome_xaxis_lims(
        ax,
        chromosome=chromosome,
        start=start,
        end=end,
    )

    if squashy:
        setup_squash_yaxis(ax)

    if chromosome is not None:
        ax.set_xlabel(f'Chromosome {chromosome}')

    else:
        ax.set_xlabel('Chromosome')

    ax.spines[['right', 'top']].set_visible(False)
    ax.xaxis.tick_bottom()
    ax.yaxis.tick_left()

    return ax


def plot_cn_profile(
        adata: AnnData,
        obs_id: str,
        value_layer_name=None,
        state_layer_name=None,
        ax=None,
        palette=None,
        chromosome=None,
        start=None,
        end=None,
        squashy=False,
        tick_major_spacing=None,
        tick_minor_spacing=None,
        **kwargs
):
    """Plot scatter points of copy number across the genome or a chromosome.

    Parameters
    ----------
    adata : AnnData
        copy number data
    obs_id : str
        observation to plot
    value_layer_name : str, optional
        layer with values for y axis, None for X, by default None
    state_layer_name : str, optional
        layer with states for colors, None for no color by state, by default None
    ax : matplotlib.axes.Axes, optional
        existing axess to plot into, by default None
    palette : str, optional
        color palette passed to sns.scatterplot
    chromosome : str, optional
        single chromosome plot, by default None
    start : int, optional
        start of plotting region
    end : int, optional
        end of plotting region
    squashy : bool, optional
        compress y axis, by default False
    rawy : bool, optional
        raw data on y axis, by default False
    tick_major_spacing : int, optional
        major tick spacing, by default 
    tick_minor_spacing : int, optional
        minor tick spacing, by default 1e6
    **kwargs :
        kwargs for sns.scatterplot

    Returns
    -------
    matplotlib.axes.Axes
        Axes used for plotting

    Examples
    -------

    .. plot::
        :context: close-figs

        import scgenome
        adata = scgenome.datasets.OV2295_HMMCopy_reduced()
        scgenome.pl.plot_cn_profile(adata, 'SA922-A90554B-R27-C43', value_layer_name='copy', state_layer_name='state')

    """

    if value_layer_name is None:
        value_layer_name = '_X'

    layers = {value_layer_name}

    if state_layer_name is not None:
        layers.add(state_layer_name)

    cn_data = get_obs_data(
        adata,
        obs_id,
        ['chr', 'start'],
        layer_names=layers)

    return plot_profile(
        cn_data,
        y=value_layer_name,
        hue=state_layer_name,
        ax=ax,
        palette=palette,
        chromosome=chromosome,
        start=start,
        end=end,
        squashy=squashy,
        tick_major_spacing=tick_major_spacing,
        tick_minor_spacing=tick_minor_spacing,
        **kwargs)



import random

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy
import scipy.cluster.hierarchy as sch
import scipy.spatial.distance as dst
import seaborn as sns
from matplotlib.colors import ListedColormap
from scgenome import refgenome
from sklearn.decomposition import PCA
from scgenome.plotting.cn import plot_profile


def hex_to_rgb(h):
    if h is None:
        return np.array((0, 0, 0))
    h = h.lstrip('#')
    return np.array(tuple(np.uint8(int(h[i:i + 2], 16)) for i in (0, 2, 4)))


color_reference = {0: '#3182BD', 1: '#9ECAE1', 2: '#CCCCCC', 3: '#FDCC8A', 4: '#FC8D59', 5: '#E34A33', 6: '#B30000',
                   7: '#980043', 8: '#DD1C77', 9: '#DF65B0', 10: '#C994C7', 11: '#D4B9DA'}


def get_cn_cmap(cn_data):
    min_cn = int(cn_data.min())
    max_cn = int(cn_data.max())
    assert min_cn - cn_data.min() == 0
    assert max_cn - cn_data.max() == 0
    color_list = []
    for cn in range(min_cn, max_cn + 1):
        if cn > max(color_reference.keys()):
            cn = max(color_reference.keys())
        color_list.append(color_reference[cn])
    return ListedColormap(color_list)


def plot_cbar(ax):
    ax.imshow(np.array([np.arange(len(color_reference))]).T[::-1], cmap=cmap, aspect=1)
    ax.set_xticks([])
    ax.set_yticks(np.arange(len(color_reference)))
    ax.set_yticklabels(np.arange(len(color_reference))[::-1])


def _secondary_clustering(data):
    D = dst.squareform(dst.pdist(data.T, 'cityblock'))
    Y = sch.linkage(D, method='complete')
    Z = sch.dendrogram(Y, color_threshold=-1, no_plot=True)
    idx = np.array(Z['leaves'])
    ordering = np.zeros(idx.shape[0], dtype=int)
    ordering[idx] = np.arange(idx.shape[0])
    return ordering


def plot_breakends(ax, breakends, lw=0.5):
    """ Plot breakpoint flags and arcs on a genome axis

    Args:
        ax: matplotlib axis
        breakends: breakend table
    
    Kwargs:
        lw: line width for arcs

    The breakends table should have the following columns:
        - prediction_id
        - chromosome
        - position
        - strand

    """
    chromosome_info = refgenome.info.chromosome_info[['chr', 'chromosome_start']].copy()
    chromosome_info['chromosome'] = chromosome_info['chr']

    plot_data = breakends.merge(chromosome_info)
    plot_data = plot_data[plot_data['chr'].isin(refgenome.info.chromosomes)]
    plot_data['plot_position'] = plot_data['position'] + plot_data['chromosome_start']

    xlim = ax.get_xlim()
    xrng = xlim[1] - xlim[0]

    ylim = ax.get_ylim()
    yrng = ylim[1] - ylim[0]

    arrow_length = 0.01 * xrng

    head_width = yrng / 30.
    head_length = xrng / 100.

    prediction_ids = list(plot_data['prediction_id'].unique())
    random.shuffle(prediction_ids)
    color_palette = sns.color_palette('hls', len(prediction_ids))
    prediction_colors = dict(zip(prediction_ids, color_palette))

    for idx in plot_data.index:
        prediction_id = plot_data.loc[idx, 'prediction_id']
        x = plot_data.loc[idx, 'plot_position']
        strand = plot_data.loc[idx, 'strand']
        y = np.random.uniform(ylim[0] + 0.75 * yrng, ylim[0] + 0.95 * yrng)
        offset = (arrow_length, -arrow_length)[strand == '+']
        color = prediction_colors[prediction_id]
        ax.arrow(x, y, offset, 0, color=color, lw=lw, alpha=1.0, head_width=head_width, head_length=head_length)
        ax.plot([x, x], [ylim[0], ylim[1]], color=color, lw=lw, ls='-', alpha=1.0)

    for prediction_id, pred_breakends in plot_data.groupby('prediction_id'):
        if len(pred_breakends.index) != 2:
            continue
        pos1 = pred_breakends['plot_position'].min()
        pos2 = pred_breakends['plot_position'].max()
        posmid = (pos1 + pos2) / 2.
        height = 0.5 * yrng * (pos2 - pos1) / xrng
        ypos = ylim[1]
        yposmid = ypos + height
        spl = scipy.interpolate.make_interp_spline([pos1, posmid, pos2], [ypos, yposmid, ypos], k=2)
        pos = np.linspace(pos1, pos2, 100)
        color = prediction_colors[prediction_id]
        ax.plot(pos, spl(pos), ls='-', lw=lw, color=color)

    ax.set_ylim((ylim[0], ylim[1] + 5))


def compute_pca_loadings(cn_data):
    """ Compute the first n components of a PCA
    """
    cn_matrix = (
        cn_data
            .set_index(['cell_id', 'chr', 'start', 'end'])['copy']
            .unstack(level=[1, 2, 3]))

    num_null = cn_matrix.isnull().sum(axis=1)
    cn_matrix = cn_matrix[num_null <= 800]
    cn_matrix = cn_matrix.dropna(axis='columns')

    pca = PCA()
    pca.fit(cn_matrix.values)

    components = pd.DataFrame(
        pca.components_,
        columns=cn_matrix.columns)

    return components


def plot_pca_components(cn_data, n_components=4, plots_prefix=None):
    """ Plot the first n components of a PCA
    """
    components = compute_pca_loadings(cn_data)

    fig = plt.figure(figsize=(20, 4 * n_components))
    for idx in range(4):
        ax = fig.add_subplot(n_components, 1, idx + 1)
        plot_data = components.iloc[idx].T.rename('component').reset_index()
        plot_profile(
            plot_data,
            ax=ax,
            s=5,
            y='component',
        )
        ax.set_ylabel(f'PCA {idx+1}')

    if plots_prefix is not None:
        fig.savefig(plots_prefix + 'pca_components.pdf', bbox_inches='tight')




def test_imports():
    import scgenome
    import scgenome.pl
    import scgenome.pp
    import scgenome.tl
    import scgenome.cnplot
    import scgenome.cncluster
    import scgenome.refgenome
    import scgenome.plotting.heatmap
    import scgenome.preprocessing.transform
    import scgenome.preprocessing.load_cn
    import scgenome.utils
    import scgenome.tools.ranges
    import scgenome.tools.getters

    print(scgenome.pl.plot_cn_profile)
    print(scgenome.pl.plot_profile)
    print(scgenome.pl.plot_cell_cn_matrix)
    print(scgenome.pl.plot_cell_cn_matrix_fig)
    print(scgenome.pl.cn_legend)
    print(scgenome.pl.plot_gc_reads)
    print(scgenome.pl.plot_tree_cn)


    print(scgenome.pp.filter_cells)
    print(scgenome.pp.calculate_filter_metrics)
    print(scgenome.pp.create_cn_anndata)
    print(scgenome.pp.read_dlp_hmmcopy)
    print(scgenome.pp.convert_dlp_hmmcopy)
    print(scgenome.pp.convert_dlp_signals)
    print(scgenome.pp.read_bam_bin_counts)
    print(scgenome.pp.read_medicc2_cn)
    print(scgenome.pp.read_snv_genotyping)


    print(scgenome.tl.cluster_cells)
    print(scgenome.tl.detect_outliers)
    print(scgenome.tl.aggregate_clusters_hmmcopy)
    print(scgenome.tl.aggregate_clusters)
    print(scgenome.tl.compute_umap)
    print(scgenome.tl.pca_loadings)
    print(scgenome.tl.sort_cells)
    print(scgenome.tl.sort_clusters)
    print(scgenome.tl.count_gc)
    print(scgenome.tl.mean_from_bigwig)
    print(scgenome.tl.add_cyto_giemsa_stain)
    print(scgenome.tl.read_ensemble_genes_gtf)
    print(scgenome.tl.aggregate_genes)
    print(scgenome.tl.ad_concat_cells)
    print(scgenome.tl.prune_leaves)
    print(scgenome.tl.align_cn_tree)
    print(scgenome.tl.create_bins)
    print(scgenome.tl.rebin)
    print(scgenome.tl.rebin_regular)
    print(scgenome.tl.weighted_mean)
    print(scgenome.tl.bin_width_weighted_mean)
    print(scgenome.tl.get_obs_data)




import pandas as pd
import anndata as ad
from scipy.sparse import csr_matrix

from anndata import AnnData


def read_snv_genotyping(filename: str) -> AnnData:
    """ Read SNV genotyping into an AnnData

    Parameters
    ----------
    filename : str
        SNV genotyping filename

    Returns
    -------
    AnnData
        SNV matrix with alt_count in X and ref_count in layers['ref_count']
    """    

    data = pd.read_csv(filename, dtype={
        'chromosome': 'category',
        'ref': 'category',
        'alt': 'category',
        'cell_id': 'category'})

    data['snv_idx'] = data.groupby(['chromosome', 'position', 'ref', 'alt'], observed=True).ngroup()

    alt_counts_matrix = csr_matrix(
        (data['alt_count'], (data['cell_id'].cat.codes, data['snv_idx'].values)),
        shape=(data['cell_id'].cat.categories.size, data['snv_idx'].max() + 1))

    ref_counts_matrix = csr_matrix(
        (data['ref_count'], (data['cell_id'].cat.codes, data['snv_idx'].values)),
        shape=(data['cell_id'].cat.categories.size, data['snv_idx'].max() + 1))

    obs = data[['cell_id']].drop_duplicates()
    obs['cell_idx'] = obs['cell_id'].cat.codes
    obs = obs.sort_values('cell_idx').set_index('cell_id').drop('cell_idx', axis=1)

    var = data[['snv_idx', 'chromosome', 'position', 'ref', 'alt']].drop_duplicates()
    var['snv_id'] = (
        var['chromosome'].astype(str) + '_' +
        var['position'].astype(str) + '_' +
        var['ref'].astype(str) + '_' +
        var['alt'].astype(str))
    var = var.sort_values('snv_idx').set_index('snv_id')
    assert (var['snv_idx'].values == range(len(var.index))).all()

    adata = ad.AnnData(
        alt_counts_matrix,
        obs,
        var,
        layers={
            'ref_count': ref_counts_matrix,
        }
    )
    
    return adata



import Bio.Phylo
import io

import scgenome.tl


def test_prune_leaves():
    # Create a small biopython tree
    tree = Bio.Phylo.read(io.StringIO('(((A:1,B:1):1,C:2):1,(D:1,E:1):3);'), 'newick')

    # Prune the leaves
    tree = scgenome.tl.prune_leaves(tree, lambda a: a.name == 'D' or a.name == 'E')

    # Verify the tree structure
    assert tree.count_terminals() == 3
    assert len(tree.clade.clades) == 1
    assert tree.clade.clades[0].clades[0].clades[0].name == 'A'
    assert tree.clade.clades[0].clades[0].clades[1].name == 'B'
    assert tree.clade.clades[0].clades[1].name == 'C'


def test_aggregate_branches():
    # Create a small biopython tree with a chain of several branches
    tree = Bio.Phylo.read(io.StringIO('(((A:1,B:1):1,C:2):1,((D:1):3):2);'), 'newick')

    print(tree)
    tree = scgenome.tl.aggregate_tree_branches(tree)

    # Verify the tree structure
    assert tree.count_terminals() == 4
    assert tree.clade.clades[0].clades[0].clades[0].name == 'A'
    assert tree.clade.clades[0].clades[0].clades[1].name == 'B'
    assert tree.clade.clades[0].clades[1].name == 'C'
    assert tree.clade.clades[1].branch_length == 6.0


def test_aggregate_branches_f_1():
    # Create a small biopython tree with a chain of several branches
    tree = Bio.Phylo.read(io.StringIO('(((A:1,B:1):1,C:2):1,((D:1):3):2);'), 'newick')

    # Add boolean wgd attribute to the tree
    tree.clade.clades[1].wgd = False
    tree.clade.clades[1].clades[0].wgd = True
    tree.clade.clades[1].clades[0].clades[0].wgd = False

    def f_merge(parent, child):
        return {
            'name': child.name,
            'branch_length': parent.branch_length + child.branch_length,
            'wgd': parent.wgd or child.wgd,
        }
    
    tree = scgenome.tl.aggregate_tree_branches(tree, f_merge)

    # Verify the tree structure
    assert tree.count_terminals() == 4
    assert tree.clade.clades[0].clades[0].clades[0].name == 'A'
    assert tree.clade.clades[0].clades[0].clades[1].name == 'B'
    assert tree.clade.clades[0].clades[1].name == 'C'
    assert tree.clade.clades[1].branch_length == 6.0
    assert tree.clade.clades[1].name == 'D'
    assert tree.clade.clades[1].wgd == True


def test_aggregate_branches_f_2():
    # Create a small biopython tree with a chain of several branches
    tree = Bio.Phylo.read(io.StringIO('Z(Y(X(D:1):3):2):4;'), 'newick')

    # Add boolean wgd attribute to the tree
    tree.clade.wgd = False
    tree.clade.clades[0].wgd = False
    tree.clade.clades[0].clades[0].wgd = True
    tree.clade.clades[0].clades[0].clades[0].wgd = False

    def f_merge(parent, child):
        return {
            'name': child.name,
            'branch_length': parent.branch_length + child.branch_length,
            'wgd': parent.wgd or child.wgd,
        }

    tree = scgenome.tl.aggregate_tree_branches(tree, f_merge)

    # Verify the tree structure
    assert tree.count_terminals() == 1
    assert tree.clade.branch_length == 10.0
    assert tree.clade.name == 'D'
    assert tree.clade.wgd == True



import numpy as np

from matplotlib.patches import Patch
from numpy import ndarray
from collections import defaultdict

color_reference = defaultdict(
    lambda: '#D4B9DA',
    {0: '#3182BD', 1: '#9ECAE1', 2: '#CCCCCC', 3: '#FDCC8A', 4: '#FC8D59', 5: '#E34A33',
     6: '#B30000', 7: '#980043', 8: '#DD1C77', 9: '#DF65B0', 10: '#C994C7', 11: '#D4B9DA'}
)


def hex_to_rgb(h):
    if h is None:
        return np.array((0, 0, 0), dtype=int)
    h = h.lstrip('#')
    return np.array(tuple(np.uint8(int(h[i:i+2], 16)) for i in (0, 2 ,4)), dtype=int)


def map_cn_colors(X: ndarray) -> ndarray:
    """ Create an array of colors from an array of copy number states

    Parameters
    ----------
    X : ndarray
        copy number states

    Returns
    -------
    ndarray
        colors with shape X.shape + (3,)
    """
    X_colors = np.zeros(X.shape + (3,), dtype=int)
    X_colors[X < 0, :] = 0
    X_colors[X > max(color_reference.keys()), :] = max(color_reference.keys())
    for state, hex in color_reference.items():
        X_colors[X == state, :] = hex_to_rgb(hex)
    return X_colors


def cn_legend(ax, frameon=True, loc=2, bbox_to_anchor=(0., 1.), title='Copy Number'):
    """ Display a legend for copy number state colors

    Parameters
    ----------
    ax : Axes
        matplotlib Axes on which to show legend
    frameon : bool, optional
        show frame, by default True
    loc : int, optional
        location of the legend, by default 2
    bbox_to_anchor : tuple, optional
        bounding box to which to anchor legend location, by default (0., 1.)

    Returns
    -------
    Legend
        legend object
    """
    states = []
    patches = []
    for s, h in color_reference.items():
        states.append(s)
        patches.append(Patch(facecolor=h, edgecolor=h))

    ncol = min(3, int(len(states)**(1/2)))

    legend = ax.legend(patches, states, ncol=ncol,
        frameon=frameon, loc=loc, bbox_to_anchor=bbox_to_anchor,
        facecolor='white', edgecolor='white', fontsize='4',
        title=title, title_fontsize='6')
    legend.set_zorder(level=200)

    return legend





from .cn import plot_cn_profile, plot_profile
from .cn_colors import cn_legend
from .heatmap import plot_cell_cn_matrix, plot_cell_cn_matrix_fig
from .qc import plot_gc_reads
from .phylo import plot_tree_cn



from anndata import AnnData
from pandas import DataFrame


def get_obs_data(
        adata: AnnData,
        obs_id: str,
        var_columns=None,
        layer_names=None) -> DataFrame:
    """ Get a dataframe for an observation including .var and specified layers

    Parameters
    ----------
    adata : AnnData
        anndata from which to retrieve observation data.
    obs_id : str
        Observation ID
    var_columns : list, optional
        List of var columns to add to returned dataframe, by default None, all columns
    layer_names : list, optional
        List of layer names to add to returned dataframe, by default None, all layers

    Returns
    -------
    DataFrame
        dataframe for an observation
    """
    if var_columns is not None and layer_names is not None and set(var_columns).intersection(layer_names):
        raise ValueError('var_columns and layer_names cannot have overlapping columns')

    elif var_columns is None and layer_names is not None:
        var_columns = list(set(adata.var.columns) - set(layer_names))

    elif layer_names is None and var_columns is not None:
        layer_names = list(set(adata.layers.keys()) - set(var_columns)) + ['_X']

    elif var_columns is None and layer_names is None:
        var_columns = adata.var.columns
        layer_names = list(adata.layers.keys()) + ['_X']

    data = adata.var[var_columns]

    for layer_name in layer_names:
        if layer_name is None or layer_name == '_X':
            layer_data = adata[obs_id].to_df().T
            assert len(layer_data.columns) == 1
            layer_data.columns = ['_X']
        else:
            layer_data = adata[obs_id].to_df(layer_name).T
            assert len(layer_data.columns) == 1
            layer_data.columns = [layer_name]

        data = data.merge(layer_data, left_index=True, right_index=True, how='left', suffixes=('', '_layer'))
    
    return data


def get_var_data(
        adata: AnnData,
        var_id: str,
        obs_columns=None,
        layer_names=None) -> DataFrame:
    """ Get a dataframe for an observation including .var and specified layers

    Parameters
    ----------
    adata : AnnData
        anndata from which to retrieve observation data.
    var_id : str
        Var ID
    obs_columns : list, optional
        List of obs columns to add to returned dataframe, by default None, all columns
    layer_names : list, optional
        List of layer names to add to returned dataframe, by default None, all layers

    Returns
    -------
    DataFrame
        dataframe for an observation
    """
    data = adata.obs

    if obs_columns is not None:
        data = data[obs_columns]

    if layer_names is None:
        layer_names = adata.layers.keys()

    for layer_name in layer_names:
        assert layer_name not in data, f'layer {layer_name} also in .obs'
        layer_data = adata[:, var_id].to_df(layer_name)
        assert len(layer_data.columns) == 1
        if layer_name is not None:
            layer_data.columns = [layer_name]
        else:
            layer_data.columns = ['_X']
        data = data.merge(layer_data, left_index=True, right_index=True, how='left')

    return data



import numpy as np
import scipy.spatial


def compute_bic(kmeans, X):
    """ Computes the BIC metric for a given k means clustering

    Args:
        kmeans: a fitted kmeans clustering object
        X: data for which to calculate bic
    
    Returns:
        float: bic
    
    Reference: https://stats.stackexchange.com/questions/90769/using-bic-to-estimate-the-number-of-k-in-kmeans
    """
    centers = [kmeans.cluster_centers_]
    labels = kmeans.labels_
    n_clusters = kmeans.n_clusters
    cluster_sizes = np.bincount(labels)
    N, d = X.shape

    # Compute variance for all clusters
    cl_var = (1.0 / (N - n_clusters) / d) * sum(
        [sum(scipy.spatial.distance.cdist(X[np.where(labels == i)], [centers[0][i]],
                                          'euclidean') ** 2) for i in range(n_clusters)])

    const_term = 0.5 * n_clusters * np.log(N) * (d + 1)

    bic = np.sum([cluster_sizes[i] * np.log(cluster_sizes[i]) -
                  cluster_sizes[i] * np.log(N) -
                  ((cluster_sizes[i] * d) / 2) * np.log(2 * np.pi * cl_var) -
                  ((cluster_sizes[i] - 1) * d / 2) for i in range(n_clusters)]) - const_term

    return bic



import seaborn as sns

import scgenome.tools.getters


def plot_gc_reads(adata, obs_id, **kwargs):
    """
     Plot scatter points of gc by read count.

    Args:
        adata (anndata.AnnData): copy number
        obs_id (str): cell or clone to plot
    
    KwArgs:
        kwargs: passed to sns.scatterplot
    """

    data = scgenome.tools.getters.get_obs_data(adata, obs_id, var_columns=['gc'])
    data = data.rename(columns={'_X': 'reads'})
    data = data[data['gc'] > 0]

    sns.scatterplot(x='gc', y='reads', data=data, **kwargs)



import unittest
import pandas as pd
from scgenome.tools.ranges import intersect_regions, intersect_positions, NoIntersectionError


class TestRanges(unittest.TestCase):

    def test_intersect_regions(self):
        # Create sample dataframes
        df1 = pd.DataFrame({
            'id': [1, 2, 3],
            'chr': ['chr1', 'chr1', 'chr2'],
            'start': [100, 200, 300],
            'end': [150, 250, 350]
        })

        df2 = pd.DataFrame({
            'id': [4, 5, 6],
            'chr': ['chr1', 'chr1', 'chr2'],
            'start': [120, 220, 320],
            'end': [170, 270, 370]
        })

        # Expected result
        expected_result = pd.DataFrame({
            'chr': ['chr1', 'chr1', 'chr2'],
            'start': [120, 220, 320],
            'end': [150, 250, 350],
            'id_x': [1, 2, 3],
            'id_y': [4, 5, 6]
        })

        # Test intersect_regions function
        result = intersect_regions(df1, df2)

        # Dont test categories or column order
        result['chr'] = result['chr'].astype(str)
        result = result[expected_result.columns]

        pd.testing.assert_frame_equal(result, expected_result)


    def test_intersect_regions_no_intersection(self):
        # Create sample dataframes with no intersection
        df1 = pd.DataFrame({
            'id': [1, 2, 3],
            'chr': ['chr1', 'chr1', 'chr2'],
            'start': [100, 200, 300],
            'end': [150, 250, 350]
        })

        df2 = pd.DataFrame({
            'id': [4, 5, 6],
            'chr': ['chr3', 'chr3', 'chr4'],
            'start': [120, 220, 320],
            'end': [170, 270, 370]
        })

        # Test intersect_regions function raises NoIntersectionError
        with self.assertRaises(NoIntersectionError):
            intersect_regions(df1, df2)


    def test_intersect_positions(self):
        # Create sample dataframes
        regions = pd.DataFrame({
            'id': [1, 2, 3],
            'chr': ['chr1', 'chr1', 'chr2'],
            'start': [100, 200, 300],
            'end': [150, 250, 350]
        })

        positions = pd.DataFrame({
            'id': [4, 5, 6, 7],
            'chr': ['chr1', 'chr1', 'chr2', 'chr2'],
            'position': [120, 180, 320, 360]
        })

        # Expected result
        expected_result = pd.DataFrame({
            'chr': ['chr1', 'chr2'],
            'position': [120, 320],
            'id_x': [1, 3],
            'id_y': [4, 6]
        })

        # Test intersect_positions function
        result = intersect_positions(regions, positions)

       # Dont test categories or column order
        result['chr'] = result['chr'].astype(str)
        result = result[expected_result.columns]

        pd.testing.assert_frame_equal(result, expected_result)


    def test_intersect_positions_no_intersection(self):
        # Create sample dataframes with no intersection
        regions = pd.DataFrame({
            'id': [1, 2, 3],
            'chr': ['chr1', 'chr1', 'chr2'],
            'start': [100, 200, 300],
            'end': [150, 250, 350]
        })

        positions = pd.DataFrame({
            'id': [4, 5, 6, 7],
            'chr': ['chr3', 'chr3', 'chr4', 'chr4'],
            'position': [120, 180, 320, 360]
        })

        # Test intersect_positions function raises NoIntersectionError
        with self.assertRaises(NoIntersectionError):
            intersect_positions(regions, positions)



if __name__ == '__main__':
    unittest.main()